{-# LANGUAGE ExplicitNamespaces, FlexibleInstances, MultiParamTypeClasses, TypeOperators, UndecidableInstances #-}
module GEODE.Metadata.Types
  ( Has(..)
  , ToJSONObject(..)
  , type (@)(..) ) where

import Data.Aeson (Object, Series, ToJSON(..), Value(..), pairs)
import Data.Aeson.KeyMap as Object (union)
import Data.Csv (FromNamedRecord(..), ToNamedRecord(..))
import Data.HashMap.Strict as Hash (union)

infixr 9 @
infixr 9 :@:
data a @ b = a :@: b

class Has a b where
  get :: b -> a

instance Has a a where
  get = id

instance Has a c => Has a (b @ c) where
  get (_ :@: c) = get c

instance {-# OVERLAPS #-} Has a (a @ b) where
  get (a :@: _) = a

instance (ToNamedRecord a, ToNamedRecord b) => ToNamedRecord (a @ b) where
  toNamedRecord (a :@: b) = Hash.union (toNamedRecord a) (toNamedRecord b)

instance (FromNamedRecord a, FromNamedRecord b) => FromNamedRecord (a @ b) where
  parseNamedRecord nr = (:@:) <$> parseNamedRecord nr <*> parseNamedRecord nr

class ToJSONObject a where
  toJSONObject :: a -> Object
  toJSONPairs :: a -> Series

instance {-# OVERLAPPABLE #-} ToJSONObject a => ToJSON a where
  toJSON = Object . toJSONObject
  toEncoding = pairs . toJSONPairs

instance (ToJSONObject a, ToJSONObject b) => ToJSONObject (a @ b) where
  toJSONObject (a :@: b) = toJSONObject a `Object.union` toJSONObject b
  toJSONPairs (a :@: b) = toJSONPairs a <> toJSONPairs b
